%{
import java.io.*;
public class Parser {
%}

%left <String> '+' '-'
%left <String> '*' '/'
%left error
%token <String> ID

%type <String> Expr

%start Expr

%%
Expr
    : Expr '+' Expr { $$ = "(" + $1 + $2 + $3 + ")"; }
    | Expr '-' Expr { $$ = "(" + $1 + $2 + $3 + ")"; }
    | Expr '*' Expr { $$ = "(" + $1 + $2 + $3 + ")"; }
    | Expr '/' Expr { $$ = "(" + $1 + $2 + $3 + ")"; }
    | Expr error { $$ = $1; }
    | '(' Expr ')' { $$ = $2; }
	| ID
    ;

%%

public Parser(Reader src) {
    reader = new LineNumberReader(src);
}

// Lex support

/** Stream from which lex reads input.
 * TODO: make this more flexible (e.g., input from a string).
 */
private LineNumberReader reader;

/** The current input line. */
private String lineStr;

/** A version of lineStr unpacked into an array for speedy access */
private char[] line = new char[100];

/** The number of characters in the current input line. */
private int lineLen;

/** The current position in the current input line (0 = start of line) */
private int pos;

/** The position in the current line where the current token started. */
private int tokenStart;

/** The next input character to be consumed.  */
private int curChar;

public int nextToken() {
    if (lookahead < 0)
        lookahead = lex();
    return lookahead;
}

public int getNextToken() {
    if (lookahead < 0)
        lookahead = lex();
    int result = lookahead;
	lookahead = -1;
	return result;
}

private final boolean getLine() {
    try {
        lineStr = reader.readLine();
    }
    catch (IOException e) {
        e.printStackTrace();
        return false;
    }
    if (lineStr == null) {
        return false;
    }

    lineLen = lineStr.length();
    
    if (line.length < lineLen + 1) {
        line = new char[lineLen + 20];
    }
    lineStr.getChars(0,lineLen,line,0);
    pos = 0;
    line[lineLen] = ' ';
    return true;
}

int lex() {
    for (;;) {
        if (pos == lineLen) {
            if (!getLine())
                return 0;
        }
        if (!Character.isWhitespace(line[pos]))
            break;
        pos++;
    }
    tokenStart = pos;

    if (Character.isJavaIdentifierStart(line[pos])) {
        do {
            pos++;
        } while (Character.isJavaIdentifierPart(line[pos]));
        String id = (new String(line, tokenStart, pos - tokenStart)).intern();
        lval = id.intern();
        return ID;
    }
    
    lval = new String(line, pos, 1);
    return line[pos++];
}

void error(String msg) {
    System.err.println(msg
            + " at line " + reader.getLineNumber()
            + " column " + tokenStart
            + ":");
    if (lineStr != null) {
        System.err.println(lineStr);
        for (int i=0; i<tokenStart; i++)
            System.err.print(lineStr.charAt(i) == '\t' ? '\t' : ' ');
        System.err.println("^");
    }
    int[] valid = validTokens(state);
	if (valid.length > 0) {
		System.err.print("Expecting ");
		for (int i=0; i<valid.length; i++) {
			if (i > 0)
				System.err.print(i == valid.length - 1 ? ", or " : ", ");
			System.err.print(tokenName(valid[i]));
		}
		System.err.println(", but found " + tokenName(lookahead));
	}
	else {
		System.err.println("No valid lookahead for state "+state
			+"; next token is " + tokenName(lookahead));
	}
}
} // Parser
