// $Header: /p/condor/repository/CONDOR_SRC/src_java/condor/classad/ClassAdWriter.java,v 1.2 2002/07/20 21:21:28 roy Exp $

/*--------------------------------------------------------------------
 *
 * Condor Java ClassAd library
 * Copyright (C) 1990-2001, CONDOR Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and Marvin Solomon
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of version 2.1 of the GNU Lesser General
 * Public License as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *--------------------------------------------------------------------*/

package condor.classad;

import java.io.*;
import java.util.Iterator;

public class ClassAdWriter {
    private static String VERSION = "$Id: ClassAdWriter.java,v 1.2 2002/07/20 21:21:28 roy Exp $";

    private PrintWriter o;

    private int indentLevel = 0;

    private void indent() {
        for (int i = 0; i < indentLevel; i++) {
            o.print("    ");
        }
    } // indent

    public ClassAdWriter(OutputStream str) {
        o = new PrintWriter(str);
    } // ClassAdWriter(OutputStream)

    public void write(Expr exp) {
        Constant c = exp.isConstant() ? (Constant) exp : null;
        switch (exp.type) {

        // Internal nodes in tree

        case Expr.ATTRIBUTE: {
            /*
            AttrRef e = (AttrRef) exp;
            if (e.locality == AttrRef.ROOT) {
                o.print('.');
            }
            */
            break;
        }
        case Expr.CALL: {
            FuncCall e = (FuncCall) exp;
            Expr[] args = e.args;
            o.print(e.func);
            o.print('(');
            for (int i = 0; i < args.length; i++) {
                if (i > 0) {
                    o.print(',');
                }
                write(args[i]);
            }
            o.print(')');
            break;
        }
        case Expr.COND: {
            CondExpr e = (CondExpr) exp;
            if (e.ec.prec() <= e.prec()) {
                o.print('(');
                write(e.ec);
                o.print(')');
            } else {
                write(e.ec);
            }
            o.print('?');
            o.print(e.et);
            o.print(':');
            o.print(e.ef);
            break;
        }
        case Expr.LIST: {
            ListExpr e = (ListExpr) exp;
            o.println('{');
            indentLevel++;
            indent();
            for (int i = 0; i < e.size(); i++) {
                if (i > 0) {
                    o.println(',');
                    indent();
                }
                write(e.sub(i));
            }
            o.println();
            indentLevel--;
            indent();
            o.print('}');
            break;
        }
        case Expr.OP: {
            Op e = (Op) exp;
            if (e.arg2 == null) {
                // unary operator
                o.print(Expr.opName[e.op]);
            }

            // left (or only) operand
            if (e.arg1.prec() < e.prec()) {
                o.print('(');
                write(e.arg1);
                o.print(')');
            } else {
                write(e.arg1);
            }

            if (e.arg2 != null) {
                // binary operator
                o.print(Expr.opName[e.op]);

                // right operand
                if (e.arg2.prec() <= e.prec()) {
                    o.print('(');
                    write(e.arg2);
                    o.print(')');
                } else {
                    write(e.arg2);
                }
            }
            break;
        }
        case Expr.RECORD: {
            RecordExpr e = (RecordExpr) exp;
            boolean first = true;

            o.println('[');
            indentLevel++;
            indent();
            for (Iterator i = e.attributes(); i.hasNext(); ) {
                String attr = (String) i.next();
                if (first) {
                    first = false;
                } else {
                    o.println(';');
                    indent();
                }
                o.print(attr);
                o.print(" = ");
                write(e.lookup(attr));
            }
            o.println();
            indentLevel--;
            indent();
            o.print(']');
            break;
        }
        case Expr.SELECTION: {
            SelectExpr e = (SelectExpr) exp;
            if (e.base.type != Expr.ATTRIBUTE
                || ((AttrRef) e.base).locality != AttrRef.SELF)
            {
                write(e.base);
                o.print('.');
            }
            o.print(e.selector);
            break;
        }
        case Expr.SUBSCRIPT: {
            SubscriptExpr e = (SubscriptExpr) exp;
            write(e.base);
            o.print('[');
            write(e.selector);
            o.print(']');
            break;
        }

        // Leaf nodes (constants)

        case Expr.ABSOLUTE_TIME:
        case Expr.RELATIVE_TIME:
            // FIXME
            o.print('\'');
            o.print(c.value);
            o.print('\'');
            break;
        case Expr.BOOLEAN:
            o.print(c == Constant.True ? "true" : "false");
            break;
        case Expr.ERROR:
            o.print("ERROR(");
            o.print(c.value);
            o.print(')');
            break;
        case Expr.INTEGER:
        case Expr.REAL:
            o.print(c.value);
            break;
        case Expr.STRING:
            // TODO: move quoteString here
            o.print(Constant.quoteString((String)c.value));
            break;
        case Expr.UNDEFINED:
            o.print("UNDEFINED(");
            o.print(c.value);
            o.print(')');
            break;
        } // switch (exp.type)
    } // write

    public void println() {
        o.println();
        o.flush();
    } // println

    public void flush() {
        o.flush();
    } // flush

} // ClassAdWriter
