%{
package condor.classad;
import java.io.*;
import java.util.Vector;
/** This is the internal classad parser generated by byacc.  From outside
 * this package, it should be accessed through the wrapper class ClassAdParser.
 * @see ClassAdParser
 */
class Parser {
private void pl(Object o) { System.err.println(o); }
%}
%token IDENTIFIER
%token INTEGER
%token REAL
%token STRING
%token TRUE
%token FALSE
%token ERROR
%token UNDEFINED
%token SUPER
%token PARENT
%token SELECT
%token FROM
%token WHERE

%left OP_LOR
%left OP_LAND
%left '|'
%left '^'
%left '&'
%left OP_EQ OP_NE OP_IS OP_ISNT
%left '<' '>' OP_LE OP_GE
%left OP_SHL OP_SHR OP_SHRR
%left '+' '-'
%left '*' '/' '%' 
%left UNARY

%type <Expr> Expr Expression PrefixExpr SuffixExpr Atom Literal
%type <FuncCall> Call
%type <RecordExpr> Record SelectionList FieldList SelectClause WhereClause
%type <ListExpr> List
%type <Vector> ListBody
%type <String> Identifier
%type <Query> Query

%start QueryOrExpression

%%
QueryOrExpression
    : Query
    | Expression
    ;
Query
    : SELECT SelectClause FROM Identifier WhereClause
        { $$ = new Query($2, $4, $5); }
    ;
SelectClause
    : '*' { $$ = null; }
    | SelectionList
    ;
WhereClause
    : WHERE Expression
        { $$ = new RecordExpr().insertAttribute("Requirements", $2); }
    | /* empty */ { $$ = null; }
    ;
Expression
    : Expr 
    | Expr '?' Expression ':' Expression { $$ = new CondExpr($1, $3, $5); }
    ;
Expr
    : Expr OP_LOR Expr  { $$ = new Op(Expr.OR, $1, $3); }
    | Expr OP_LAND Expr  { $$ = new Op(Expr.AND, $1, $3); }
    | Expr '|' Expr  { $$ = new Op(Expr.BITOR, $1, $3); }
    | Expr '^' Expr  { $$ = new Op(Expr.BITXOR, $1, $3); }
    | Expr '&' Expr  { $$ = new Op(Expr.BITAND, $1, $3); }
    | Expr OP_EQ Expr  { $$ = new Op(Expr.EQUAL, $1, $3); }
    | Expr OP_NE Expr  { $$ = new Op(Expr.NOT_EQUAL, $1, $3); }
    | Expr OP_IS Expr  { $$ = new Op(Expr.SAME, $1, $3); }
    | Expr OP_ISNT Expr  { $$ = new Op(Expr.DIFFERENT, $1, $3); }
    | Expr '<' Expr  { $$ = new Op(Expr.LESS, $1, $3); }
    | Expr '>' Expr  { $$ = new Op(Expr.GREATER, $1, $3); }
    | Expr OP_LE Expr  { $$ = new Op(Expr.LESS_EQ, $1, $3); }
    | Expr OP_GE Expr  { $$ = new Op(Expr.GREATER_EQ, $1, $3); }
    | Expr OP_SHL Expr  { $$ = new Op(Expr.LEFT_SHIFT, $1, $3); }
    | Expr OP_SHR Expr  { $$ = new Op(Expr.RIGHT_SHIFT, $1, $3); }
    | Expr OP_SHRR Expr  { $$ = new Op(Expr.URIGHT_SHIFT, $1, $3); }
    | Expr '+' Expr  { $$ = new Op(Expr.PLUS, $1, $3); }
    | Expr '-' Expr  { $$ = new Op(Expr.MINUS, $1, $3); }
    | Expr '*' Expr { $$ = new Op(Expr.TIMES, $1, $3); }
    | Expr '/' Expr  { $$ = new Op(Expr.DIV, $1, $3); }
    | Expr '%' Expr  { $$ = new Op(Expr.MOD, $1, $3); }
    | PrefixExpr
    ;
PrefixExpr
    : SuffixExpr
    | '+' PrefixExpr %prec UNARY { $$ = new Op(Expr.UPLUS, $2); }
    | '-' PrefixExpr %prec UNARY { $$ = new Op(Expr.UMINUS, $2); }
    | '!' PrefixExpr %prec UNARY { $$ = new Op(Expr.NOT, $2); }
    | '~' PrefixExpr %prec UNARY { $$ = new Op(Expr.BIT_COMPLEMENT, $2); }
    ;
SuffixExpr
    : Atom
    | '.' Identifier { $$ = new SelectExpr(new AttrRef(AttrRef.ROOT), $2); }
    | SuffixExpr '.' Identifier { $$ = new SelectExpr($1,  $3); }
    | SuffixExpr '[' Expression ']' { $$ = new SubscriptExpr($1, $3); }
    ;
Atom
    : Identifier { $$ = new SelectExpr(new AttrRef(AttrRef.SELF), $1); }
    | Literal
    | List { $$ = $1; }
    | Record { $$ = $1; }
    | Call { $$ = $1; }
    | '(' Expression ')' { $$ = $2; }
    ;
List
    : '{' ListBody ListEnd { $$ = new ListExpr($2); }
    | '{' ListEnd { $$ = new ListExpr(null); }
    ;
ListEnd
    : '}'
    | ',' '}'
    ;
ListBody
    : Expression { Vector v = new Vector(); v.addElement($1); $$ = v; }
    | ListBody ',' Expression {
        Vector v = $1; v.addElement($3); $$ = v;
    }
    ;
Record
    : '[' FieldList RecordEnd { $$ = $2; }
    | '[' RecordEnd { $$ = new RecordExpr(); }
    ;
RecordEnd
    : ']'
    | ';' ']'
    ;
FieldList
    : Identifier '=' Expression {
       RecordExpr r = new RecordExpr();
       r.insertAttribute($1, $3);
       $$ = r;
    }
    | FieldList ';' Identifier '=' Expression {
        RecordExpr r = $1;
        if (r.lookup($3) != null)
            error("Duplicate definition of " + $3);
        r.insertAttribute($3, $5);
        $$ = r;
    }
    ;
SelectionList
    : Identifier '=' Expression {
       RecordExpr r = new RecordExpr();
       r.insertAttribute($1, $3);
       $$ = r;
    }
    | SelectionList ',' Identifier '=' Expression {
        RecordExpr r = $1;
        if (r.lookup($3) != null)
            error("Duplicate definition of " + $3);
        r.insertAttribute($3, $5);
        $$ = r;
    }
    | Expression {
       RecordExpr r = new RecordExpr();
       r.insertAttribute($1);
       $$ = r;
    }
    | SelectionList ',' Expression {
        RecordExpr r = $1;
        r.insertAttribute($3);
        $$ = r;
    }
    ;
Call
    : Identifier '(' ListBody ')' { $$ = new FuncCall($1, $3); }
    | Identifier '(' ')' { $$ = new FuncCall($1, null); }
    ;

Literal
    : REAL { $$ = lval; }
    | STRING { $$ = lval; }
    | INTEGER { $$ = lval; }
    | TRUE { $$ = Constant.True; }
    | FALSE { $$ = Constant.False; }
    | UNDEFINED { $$ = Constant.Undef; }
    | ERROR { $$ = Constant.Error; }
    | SUPER { $$ = new AttrRef(AttrRef.SUPER); }
    | PARENT { $$ = new AttrRef(AttrRef.PARENT); }
    // Todo:  time constants
    ;
Identifier
    : IDENTIFIER { $$ = lval; }
    ;

%%
private static String VERSION = "$Id: classad.y,v 1.10 2002/03/13 15:00:15 solomon Exp $";

/** Verbosity for error messages.
 * 0 - no messages
 * 1 - only show message and line number
 * 2 - also echo line in question
 * 3 - also list tokens expected
 */
int verbosity = 3;

public Parser(Reader rdr) {
    reader = new LineNumberReader(rdr);
}

public Parser(InputStream str) {
    reader = new LineNumberReader(new InputStreamReader(str));
}

public Parser(String str) {
    reader = new LineNumberReader(new StringReader(str));
}

public void reset(Reader rdr) {
    reader = new LineNumberReader(rdr);
    lookahead = -1;
}

public void reset(InputStream str) {
    reader = new LineNumberReader(new InputStreamReader(str));
    lookahead = -1;
}

public void reset(String str) {
    reader = new LineNumberReader(new StringReader(str));
    lookahead = -1;
}

// Lex support

/** Stream from which lex reads input.
 * TODO: make this more flexible (e.g., input from a string).
 */
private LineNumberReader reader;

/** The current input line. */
private String lineStr;

/** A version of lineStr unpacked into an array for speedy access */
private char[] line = new char[100];

/** The number of characters in the current input line. */
private int lineLen;

/** The current position in the current input line (0 = start of line) */
private int pos;

/** The position in the current line where the current token started. */
private int tokenStart;

/** The next input character to be consumed.  */
private int curChar;

/** A table of reserved words. */
private static final String[] reserved = {
    null,
    "error",
    "false",
    "from",
    "is",
    "isnt",
    "select",
    "true",
    "undefined",
    "where"
};

/** A trie constructed from the reserved words.
 * Three parallel arrays.  The element match[j] is supposed to match the
 * next letter of the input string.  If accept[j] = n != 0, then the string
 * matched thus far is reserved word n.  If the match[j] does not match
 * the next input character and alt[j] = n != 0, try match[n] instead.
 * If the input string starts with letter c, then start matching at position
 * trieStart[c], provided trieStart[c] > 0.
 */
private static char[] match;
private static int[] accept;
private static int[] alt;
private static int[] trieStart;
static {
    String m = " error false rom isnt select true undefined where ";
        //      01234567890123456789012345678901234567890123456789
        //                1         2         3         4
    match = m.toCharArray();

    trieStart = new int['z'];
    trieStart['e'] = trieStart['E'] = 1;
    trieStart['f'] = trieStart['F'] = 7;
    trieStart['i'] = trieStart['I'] = 17;
    trieStart['s'] = trieStart['S'] = 22;
    trieStart['t'] = trieStart['T'] = 29;
    trieStart['u'] = trieStart['U'] = 34;
    trieStart['w'] = trieStart['W'] = 44;

    accept = new int[m.length()];
    accept[6] = 1;
    accept[12] = 2;
    accept[16] = 3;
    accept[19] = 4;
    accept[21] = 5;
    accept[28] = 6;
    accept[33] = 7;
    accept[43] = 8;
    accept[49] = 9;

    alt = new int[m.length()];
    alt[8] = 13;
};

/** A table of token codes parallel to the array reserved. */
private static final int[] reservedCode = {
    0, // unused
    ERROR,
    FALSE,
    FROM,
    OP_IS,
    OP_ISNT,
    SELECT,
    TRUE,
    UNDEFINED,
    WHERE
};

/** A "wired-in" version of Character.toLowerCase(c) to avoid the
 * overhead in the common case of ASCII (c < 128).
 */
private static final int[] tolower = new int[] {
       0,   1,   2,   3,   4,   5,   6,   7,
       8,   9,  10,  11,  12,  13,  14,  15,
      16,  17,  18,  19,  20,  21,  22,  23,
      24,  25,  26,  27,  28,  29,  30,  31,
      32,  33,  34,  35,  36,  37,  38,  39,
      40,  41,  42,  43,  44,  45,  46,  47,
      48,  49,  50,  51,  52,  53,  54,  55,
      56,  57,  58,  59,  60,  61,  62,  63,
      64,  97,  98,  99, 100, 101, 102, 103,
     104, 105, 106, 107, 108, 109, 110, 111,
     112, 113, 114, 115, 116, 117, 118, 119,
     120, 121, 122,  91,  92,  93,  94,  95,
      96,  97,  98,  99, 100, 101, 102, 103,
     104, 105, 106, 107, 108, 109, 110, 111,
     112, 113, 114, 115, 116, 117, 118, 119,
     120, 121, 122, 123, 124, 125, 126, 127
};

void lexdebug(Object msg) {
    System.out.println("LEX: line " + reader.getLineNumber()
        + " column " + (tokenStart+1)
        + " " + msg);
}

/** Get the "lookahead" token: the first token not yet consumed by the
 * parser.
 * The token is not removed from the input stream.
 * @return the code for the next token.
 * @see #getNextToken()
 */
public int nextToken() {
    if (lookahead < 0)
        lookahead = lex();
    return lookahead;
}

/** Get the value associated with the "lookahead" token: the first token not
 * yet consumed by the parser.
 * The token is not removed from the input stream.
 * @return the value associated with the next token.
 * @see #nextToken()
 */
public Object nextValue() {
    if (lookahead < 0)
        lookahead = lex();
    return lval;
}

/** Get and remove the "lookahead" token: the first token not yet consumed
 * by the parser.
 * The token is removed from the input stream.
 * @return the code for the next token.
 * @see #getNextToken()
 */
public int getNextToken() {
    if (lookahead < 0)
        lookahead = lex();
    int result = lookahead;
    lookahead = -1;
    return result;
}
/** Gets the next line from reader and unpacks it to the array line[].
 * @return -1 fo rerrors, 0 for EOF, or 1 for success.
 */
private final int getLine() {
    try {
        lineStr = reader.readLine();
    }
    catch (IOException e) {
        e.printStackTrace();
        return -1;
    }
    if (lineStr == null) {
        return 0; // EOF
    }

    lineLen = lineStr.length();
    
    if (line.length < lineLen + 1) {
        line = new char[lineLen + 20];
    }
    lineStr.getChars(0,lineLen,line,0);
    pos = 0;
    line[lineLen] = ' ';
    return 1;
}

/** Check whether the word stored in the line[] array starting at position
 * tokenStart matches one of the reserved words.  If so, return its index
 * in the array reserved[].  Otherwise, return 0.
 * @return an index in reserved[] or 0.
 */
private final int reservedLookup() {
    // Find the starting point in the trie
    int c = line[tokenStart];
    if (c >= trieStart.length) {
        return 0;
    }
    int pos = trieStart[c];
    if (pos++ == 0) {
        return 0;
    }
    // Start matching with the second character of the input
    for (int i = tokenStart+1; ;i++) {
        c = line[i];
        if (!Character.isJavaIdentifierStart((char)c)) {
            // End of input reached.  Return the code stored in the accept[]
            // array (will be zero if we are not at an "accept" node in the
            // trie.
            return accept[pos];
        }
        if (c >= tolower.length) {
            // Reserved words in the trie are entirely made up of ASCII
            // characters (<128).  If the input contains anything else, it is
            // obviously not a reserved word.
            return 0;
        }
        // Translate to lower case
        c = tolower[c];
        while (match[pos] != c) {
            // Match failed.  See if there is an alternative match at this
            // point (sibling node in the trie).
            pos = alt[pos];
            if (pos == 0) {
                // No alternative.  No match.
                return 0;
            }
        }
        // Successfully matched one more charcter of input.
        pos++;
    }
}

/** The lexical scanner.
 * It finds the next token in the input stream and returns its token code.
 * It also sets the fields tokenStart and pos so that the text of the token
 * is in line[tokenStart ... pos-1]. (For efficiency reasons, we do not
 * return the text in text, since that would entail allocation a new String
 * that would, in most cases, immediately become garbage.)  
 * The "semantics" associated with the current token (if any) is stored
 * in lval.
 * <p>
 * @return the token code for the next token; 0 for EOF; -1 for error.
 */
int lex() {
    int n;
    // Skip spaces and comments
    for (;;) {
        if (pos == lineLen || (line[pos] == '/' && line[pos+1] == '/')) {
            if ((n = getLine()) <= 0)
                return n;
            continue;
        }
        if (line[pos] == '/' && line[pos+1] == '*') {
            for (pos += 2;;pos++) {
                if (pos == lineLen) {
                    if ((n = getLine()) <= 0)
                        return n;
                    continue;
                }
                if (line[pos] == '*' && line[pos+1] == '/') {
                    break;
                }
            }
            continue;
        }
        if (!Character.isWhitespace(line[pos])) {
            break;
        }
        pos++;
    }
    tokenStart = pos;

    // Identifier or reserved word
    if (Character.isJavaIdentifierStart(line[pos])) {
        do {
            pos++;
        } while (Character.isJavaIdentifierPart(line[pos]));
        int i = reservedLookup();
        if (i > 0) {
            if (debug)
                lexdebug("reserved word " + reservedCode[i]
                    + " (" + reserved[i] + ")");
            lval = null; // ???
            return reservedCode[i];
        }
        lval = new String(line, tokenStart, pos - tokenStart);
        if (debug)
            lexdebug("ID " + lval);
        return IDENTIFIER;
    }

    // Number or '.'
    if (Character.isDigit(line[pos]) || line[pos]=='.') {
        // Number or lone '.'
        int type;
        if (line[pos] == '.' && !Character.isDigit(line[pos+1])) {
            pos++;
            if (debug)
                lexdebug("DOT");
            return '.';
        }
        // Skip any leading digits
        for (; Character.isDigit(line[pos]); pos++) {
        }
        if (line[pos] == '.') {
            // There is a decimal point.  Skip the fraction part if any.
            type = REAL;
            for (pos++; Character.isDigit(line[pos]); pos++) {
            }
        }
        else {
            // There was no decimal point.  For now, assume it is an
            // integer.
            type = INTEGER;
            if (line[tokenStart] == '0'
                && (line[tokenStart+1] == 'x' || line[tokenStart+1] == 'x'))
            {
                // A hex integer constant.  Note that 'e' and 'E' are hex
                // digits, so 0x123e12 is parsed as a hex integer, and
                // and 0x123e+12 is parsed as three tokens: 0x123e + 12
                int value = 0;
                int digitValue;
                for (pos = tokenStart+2;
                    (digitValue = Character.digit(line[pos],16)) >= 0;
                    pos++)
                {
                    value = (value << 4) + digitValue;
                }
                // There must be at least one digit
                if (pos == tokenStart + 2) {
                    if (debug)
                        lexdebug("bad number");
                    return -1;
                }
                lval = Constant.getInstance(value);
                if (debug)
                    lexdebug("INTEGER " + lval);
                return INTEGER;
            }
        }
        // Now look for an exponent
        if (line[pos] == 'e' || line[pos] == 'E') {
            type = REAL;
            pos++;
            if (line[pos] == '+' || line[pos] == '-') {
                pos++;
            }
            if (!Character.isDigit(line[pos])) {
                if (debug)
                    lexdebug("bad float");
                return -1;
            }
            for (pos++; Character.isDigit(line[pos]); pos++) {
            }
        }
        // Finally, compute the value
        if (type == REAL) {
            lval = Constant.getInstance(
                        Double.parseDouble(
                            new String(line, tokenStart, pos - tokenStart)));
            if (debug)
                lexdebug("REAL " + lval);
            return REAL;
        }
        // If there was no decimal point or exponent, we have to check
        // for a leading 0, meaning octal.
        int value = 0;
        int radix = (line[tokenStart] == '0') ? 8 : 10;
        for (int i=tokenStart; i<pos; i++) {
            int digitValue = Character.digit(line[i],radix);
            if (digitValue < 0) {
                if (debug)
                    lexdebug("bad oct");
                return -1;
            }
            value = value * radix + digitValue;
        }
        lval = Constant.getInstance(value);
        if (debug)
            lexdebug("INT " + lval);
        return INTEGER;
    } // Number

    // Multi-character operators
    switch (line[pos++]) {
        case '!':
            if (line[pos] == '=') {
                pos++;
                if (debug)
                    lexdebug("NE");
                return OP_NE;
            }
            break;
        case '&':
            if (line[pos] == '&') {
                pos++;
                if (debug)
                    lexdebug("LAND");
                return OP_LAND;
            }
            break;
        case '<':
            if (line[pos] == '<') {
                pos++;
                if (debug)
                    lexdebug("SHL");
                return OP_SHL;
            }
            if (line[pos] == '=') {
                pos++;
                if (debug)
                    lexdebug("LE");
                return OP_LE;
            }
            break;
        case '=':
            if (line[pos] == '=') {
                pos++;
                if (debug)
                    lexdebug("EQ");
                return OP_EQ;
            }
            if (line[pos] == '?' && line[pos+1] == '=') {
                pos += 2;
                if (debug)
                    lexdebug("IS");
                return OP_IS;
            }
            if (line[pos] == '!' && line[pos+1] == '=') {
                pos += 2;
                if (debug)
                    lexdebug("ISNT");
                return OP_ISNT;
            }
            break;
        case '>':
            if (line[pos] == '=') {
                pos++;
                if (debug)
                    lexdebug("GE");
                return OP_GE;
            }
            if (line[pos] == '>') {
                pos++;
                if (line[pos] == '>') {
                    pos++;
                    if (debug)
                        lexdebug("SHRR");
                    return OP_SHRR;
                }
                if (debug)
                    lexdebug("SHR");
                return OP_SHR;
            }
            break;
        case '|':
            if (line[pos] == '|') {
                pos++;
                if (debug)
                    lexdebug("LOR");
                return OP_LOR;
            }
            break;
        case '"':
            for (;;) {
                if (pos == lineLen) {
                    // Unterminated string constant.  Really should issue
                    // a warning here!
                    line[pos++] = '"';
                    break;
                }
                if (line[pos] == '\\') {
                    pos +=2 ;
                    continue;
                }
                if (line[pos] == '"') {
                    pos++;
                    break;
                }
                pos++;
            }
            lval = Constant.getInstance(line, tokenStart, pos-1);
            if (debug)
                lexdebug("STRING " + lval);
            return STRING;
    }
    // In all other cases, simply return the single character.  If it
    // isn't legit, the parser will catch it.
    if (debug)
        lexdebug("'" + line[pos-1] + "'");
    return line[pos-1];
} // lex

void error(String msg) {
    if (verbosity <= 0) {
        return;
    }
    if (verbosity == 1) {
        System.err.println(msg
                + " at line " + reader.getLineNumber()
                + " column " + tokenStart);
        return;
    }
    System.err.println(msg
            + " at line " + reader.getLineNumber()
            + " column " + tokenStart
            + ":");
    if (lineStr != null) {
        System.err.println(lineStr);
        for (int i=0; i<tokenStart; i++)
            System.err.print(lineStr.charAt(i) == '\t' ? '\t' : ' ');
        System.err.println("^");
    }
    if (verbosity < 3) {
        return;
    }
    int[] valid = validTokens(state);
    if (valid.length > 0) {
        System.err.print("Expecting ");
        for (int i=0; i<valid.length; i++) {
            if (i > 0)
                System.err.print(i == valid.length - 1 ? ", or " : ", ");
            System.err.print(tokenName(valid[i]));
        }
        System.err.println(", but found " + tokenName(lookahead));
    }
    else {
        System.err.println("No valid lookahead for state "+state
            +"; next token is " + tokenName(lookahead));
    }
} // error
} // Parser
